"""Add user.cache_dir

Revision ID: 385aa8c18489
Revises: 47072e0106ce
Create Date: 2023-04-06 10:13:43.234442

"""
import sqlalchemy as sa
import sqlmodel
from alembic import op


# revision identifiers, used by Alembic.
revision = "385aa8c18489"
down_revision = "47072e0106ce"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "user_oauth",
        sa.Column(
            "cache_dir", sqlmodel.sql.sqltypes.AutoString(), nullable=True
        ),
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("user_oauth", "cache_dir")
    # ### end Alembic commands ###
