"""Drop project.project_dir

Revision ID: 6dede8d6fd9d
Revises: 385aa8c18489
Create Date: 2023-04-06 10:32:52.972906

"""
import sqlalchemy as sa
from alembic import op


# revision identifiers, used by Alembic.
revision = "6dede8d6fd9d"
down_revision = "385aa8c18489"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("project", "project_dir")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "project", sa.Column("project_dir", sa.VARCHAR(), nullable=False)
    )
    # ### end Alembic commands ###
